import lmfit as lf
import numpy as np
import xarray as xr

from xrfit.base import DataArrayAccessor


@xr.register_dataarray_accessor("fit")
class FitAccessor(DataArrayAccessor):
    def guess(
        self,
        model: lf.model.Model,
        input_core_dims: str = "x",
    ) -> xr.DataArray:
        """
        Generate initial guess for the model parameters.

        model : lf.model.Model
            The model for which to generate the initial guess.
        input_core_dims : str, optional
            The dimension name in the xarray object to be used as input for the model's guess function. Default is "x".

        Returns
        -------
        xr.DataArray
            An xarray DataArray containing the initial guess for the model parameters.

        Notes
        -----
        This method uses `xr.apply_ufunc` to apply the model's guess function to the data
        """
        return xr.apply_ufunc(
            model.guess,
            self._obj,
            input_core_dims=[[input_core_dims]],
            kwargs={
                "x": getattr(self._obj, input_core_dims).values,
            },
            vectorize=True,
            dask="parallelized",
            output_dtypes=[object],
        )

    def _update(
        self,
        params: xr.DataArray,
        params_new: xr.DataArray,
    ) -> xr.DataArray:
        """
        Update the parameters with new values.

        This method takes two xarray DataArray objects, `params` and `params_new`,
        and updates the values in `params` with the corresponding values from
        `params_new`.

        Parameters
        ----------
        params : xr.DataArray
            The original parameters to be updated.
        params_new : xr.DataArray
            The new parameters to update the original parameters with.

        Returns
        -------
        xr.DataArray
            The updated parameters as an xarray DataArray.
        """
        return xr.apply_ufunc(
            lambda x, y: x.update(y),
            params,
            params_new,
            vectorize=True,
            dask="parallelized",
            output_dtypes=[object],
        )

    def __call__(
        self,
        model: lf.model.Model,
        params: xr.DataArray | None = None,
        input_core_dims: str = "x",
        **kws,
    ) -> lf.model.ModelResult:
        """
        Call method to fit a model to the data.

        Parameters
        ----------
        model : lf.model.Model
            The model to be fitted.
        params : xr.DataArray or None, optional
            The parameters for the model. If None, parameters will be guessed.
        input_core_dims : str, optional
            The dimension name for the input data, by default "x".

        Returns
        -------
        xr.DataArray
            The result of the model fitting.

        """
        guesses = self.guess(model, input_core_dims)
        guesses = self._update(guesses, params) if params is not None else guesses

        args = [kws.pop(name) for name in ["weights"] if name in kws]
        input_core_dims_new = [
            [input_core_dims],
            [],
            *[[input_core_dims] for _ in args],
        ]
        return xr.apply_ufunc(
            model.fit,
            self._obj,
            guesses,
            *args,
            input_core_dims=input_core_dims_new,
            kwargs={
                "x": getattr(self._obj, input_core_dims).values,
                **kws,
            },
            vectorize=True,
            dask="parallelized",
        )

    # def _get_attrs(
    #     self,
    #     fit_results: xr.DataArray,

    # ):
    #     return xr.apply_ufunc(
    #         lambda x: x.,
    #         fit_results,
    #         vectorize=True,
    #         dask="parallelized",
    #         output_dtypes=[object],
    #     )

    def fit_with_corr(
        self,
        model: lf.model.Model,
        input_core_dims: str = "x",
        start_dict: dict | None = None,
        **kws,
    ) -> xr.DataArray:
        """
        Fit the model starting from a certain index and use the resulting parameters for the next fit.

        Parameters
        ----------
        model : lf.model.Model
            The model to be fitted.
        start_index : tuple
            The starting index for the fit.
        input_core_dims : str, optional
            The dimension name for the input data, by default "x".

        Returns
        -------
        xr.DataArray
            The result of the model fitting with correlated parameters.
        """
        fit_results = self.__call__(model=model, input_core_dims=input_core_dims)
        dims = fit_results.dims
        if start_dict is None:
            start_dict = fit_results.assess.best_fit_stat()
            print("⚡️ No initial coords provided for fit_with_corr")
            print("⚡️ Estimate used :", start_dict)
        start_tuple = tuple(start_dict.values())
        dims_tuple = tuple(fit_results.sizes[dim] for dim in dims)
        start_idx = np.ravel_multi_index(start_tuple, dims_tuple)
        total_idx = np.prod(dims_tuple)
        previous_params = fit_results.params.__call__().isel(start_dict).item()

        for idx in range(start_idx - 1, -1, -1):
            indices = np.unravel_index(idx, dims_tuple)
            index_dict = dict(zip(dims, indices, strict=False))
            single_fit_result = fit_results.isel(index_dict).item()
            single_fit_result.fit(params=previous_params)
            fit_results[index_dict] = single_fit_result
            previous_params = single_fit_result.params

        previous_params = fit_results.params.__call__().isel(start_dict).item()

        for idx in range(start_idx + 1, total_idx):
            indices = np.unravel_index(idx, dims_tuple)
            index_dict = dict(zip(dims, indices, strict=False))
            single_fit_result = fit_results.isel(index_dict).item()
            single_fit_result.fit(params=previous_params)
            fit_results[index_dict] = single_fit_result
            previous_params = single_fit_result.params

        return fit_results



import lmfit as lf
import numpy as np
import xarray as xr
from lmfit.models import LorentzianModel


def get_test_data_2d():
    rng = np.random.default_rng(seed=0)
    x = np.linspace(-10, 10, 200)
    y = np.linspace(-5, 5, 7)
    model = LorentzianModel()
    return xr.DataArray(
        np.stack(
            [
                model.eval(x=x, amplitude=1, center=0, sigma=0.05 * 5),
                model.eval(x=x, amplitude=1, center=0, sigma=0.1 * 5),
                model.eval(x=x, amplitude=1, center=0, sigma=0.15 * 5),
                model.eval(x=x, amplitude=1, center=0, sigma=0.2 * 5),
                model.eval(x=x, amplitude=1, center=0, sigma=0.25 * 5),
                model.eval(x=x, amplitude=1, center=0, sigma=0.3 * 5),
                model.eval(x=x, amplitude=1, center=0, sigma=0.35 * 5),
            ]
        ).T
        + rng.normal(size=(x.size, y.size)) * 0.01,
        coords={"x": x, "y": y},
        dims=("x", "y"),
    )


def get_test_data_3d():
    rng = np.random.default_rng(seed=0)
    x = np.linspace(-10, 10, 200)
    y = np.linspace(-5, 5, 2)
    z = np.linspace(0, 2, 3)
    model = LorentzianModel()
    return xr.DataArray(
        np.stack(
            [
                [
                    model.eval(x=x, amplitude=1, center=0, sigma=0.1 * 5),
                    model.eval(x=x, amplitude=1, center=0, sigma=0.2 * 5),
                ],
                [
                    model.eval(x=x, amplitude=1, center=0, sigma=0.2 * 5),
                    model.eval(x=x, amplitude=1, center=0, sigma=0.3 * 5),
                ],
                [
                    model.eval(x=x, amplitude=1, center=0, sigma=0.3 * 5),
                    model.eval(x=x, amplitude=1, center=0, sigma=0.4 * 5),
                ],
            ],
        ).transpose()
        + rng.normal(size=(x.size, y.size, z.size)) * 0.01,
        coords={"x": x, "y": y, "z": z},
        dims=("x", "y", "z"),
    )


def get_test_result(data):
    model = LorentzianModel()
    guess = data.fit.guess(model=model)
    return data.fit(model=model, params=guess)


def test_fit_3d(qtbot):
    rng = np.random.default_rng(seed=0)
    x = np.linspace(-10, 10, 200)
    y = np.linspace(-5, 5, 2)
    z = np.linspace(0, 2, 3)
    model = LorentzianModel()
    data = xr.DataArray(
        np.stack(
            [
                [
                    model.eval(x=x, amplitude=1, center=0, sigma=0.1),
                    model.eval(x=x, amplitude=1, center=0, sigma=0.2),
                ],
                [
                    model.eval(x=x, amplitude=1, center=0, sigma=0.2),
                    model.eval(x=x, amplitude=1, center=0, sigma=0.3),
                ],
                [
                    model.eval(x=x, amplitude=1, center=0, sigma=0.3),
                    model.eval(x=x, amplitude=1, center=0, sigma=0.4),
                ],
            ],
        ).transpose()
        + rng.normal(size=(x.size, y.size, z.size)) * 0.01,
        coords={"x": x, "y": y, "z": z},
        dims=("x", "y", "z"),
    )

    assert isinstance(data, xr.DataArray)
    assert data.shape == (x.size, y.size, z.size)
    assert data.dims == ("x", "y", "z")

    guess = data.fit.guess(model=model)
    assert isinstance(guess, xr.DataArray)
    assert guess.shape == (y.size, z.size)
    assert guess.dims == ("y", "z")
    assert isinstance(guess[0, 0].item(), lf.Parameters)

    result = data.fit(model=model, params=guess)
    assert isinstance(result, xr.DataArray)
    assert result.shape == (y.size, z.size)
    assert result.dims == ("y", "z")
    assert isinstance(result[0, 0].item(), lf.model.ModelResult)


def test_fit(qtbot):
    rng = np.random.default_rng(seed=0)
    x = np.linspace(-10, 10, 200)
    y = np.linspace(0, 2, 3)
    model = LorentzianModel()
    data = xr.DataArray(
        np.stack(
            [
                model.eval(x=x, amplitude=1, center=0, sigma=0.1 * 5),
                model.eval(x=x, amplitude=1, center=0, sigma=0.2 * 5),
                model.eval(x=x, amplitude=1, center=0, sigma=0.3 * 5),
            ]
        ).T
        + rng.normal(size=(x.size, y.size)) * 0.01,
        coords={"x": x, "y": y},
        dims=("x", "y"),
    )

    assert isinstance(data, xr.DataArray)
    assert data.shape == (x.size, y.size)
    assert data.dims == ("x", "y")

    guess = data.fit.guess(model=model)
    assert isinstance(guess, xr.DataArray)
    assert guess.shape == (y.size,)
    assert guess.dims == ("y",)
    assert isinstance(guess[0].item(), lf.Parameters)

    result = data.fit(model=model, params=guess)
    assert isinstance(result, xr.DataArray)
    assert result.shape == (y.size,)
    assert result.dims == ("y",)
    assert isinstance(result[0].item(), lf.model.ModelResult)

    params = result.params()
    assert isinstance(params, xr.DataArray)
    assert isinstance(params[0].item(), lf.Parameters)
    sorted_result = result.params.sort("center")
    assert isinstance(sorted_result, xr.DataArray)
    assert isinstance(sorted_result[0].item(), lf.model.ModelResult)
    smoothend_result = result.params.smoothen("center", 5)
    assert isinstance(smoothend_result, xr.DataArray)
    assert isinstance(smoothend_result[0].item(), lf.model.ModelResult)



from xrfit.arr import ArrAccessor
from xrfit.assess import AccessAccessor
from xrfit.display import DisplayAccessor
from xrfit.fit import FitAccessor
from xrfit.params import ParamsAccessor

__all__ = [
    "AccessAccessor",
    "ArrAccessor",
    "DisplayAccessor",
    "FitAccessor",
    "ParamsAccessor",
]



import numpy as np
import xarray as xr
from scipy.ndimage import gaussian_filter

from xrfit.base import DataArrayAccessor


@xr.register_dataarray_accessor("params")
class ParamsAccessor(DataArrayAccessor):
    def __call__(
        self,
    ) -> xr.DataArray:
        return xr.apply_ufunc(
            lambda x: x.params,
            self._obj,
            vectorize=True,
            dask="parallelized",
            output_dtypes=[object],
        )

    def smoothen(
        self,
        param_name: str = "center",
        sigma: int = 5,
        # params_name: list | None = None,
    ) -> xr.DataArray:
        # if params_name is None:
        # params_name = ["center"]
        param = self._obj.params.get(param_name)
        smoothing_sigma = [
            sigma if i < param.ndim - 1 else 0 for i in range(param.ndim)
        ]

        param_smooth = gaussian_filter(param, sigma=smoothing_sigma)
        self._obj.params.set(param_smooth, param_name)
        return self._obj

    def sort(
        self,
        target_param_name: str = "center",
        params_name: list | None = None,
    ) -> xr.DataArray:
        if params_name is None:
            params_name = ["center"]
        param_to_sortby = self._obj.params.get(target_param_name)
        sorted_indices = param_to_sortby.argsort(axis=-1)
        for param_name in params_name:
            param = self._obj.params.get(param_name)
            sorted_param = param.isel(params_dim=sorted_indices)
            self._obj.params.set(sorted_param, param_name)
        return self._obj

    def get(
        self,
        params_name: str = "center",
        params_attr: str = "value",
    ) -> xr.DataArray:
        return xr.apply_ufunc(
            self._get,
            self._obj,
            kwargs={
                "params_name": params_name,
                "params_attr": params_attr,
            },
            input_core_dims=[[]],
            output_core_dims=[["params_dim"]],
            vectorize=True,
        )

    def _get(
        self,
        data: xr.DataArray,
        params_name: str = "center",
        params_attr: str = "value",
    ):
        params = data.params
        return np.array(
            [
                getattr(params[key], params_attr)
                for key in params
                if key.endswith(params_name)
            ]
        )

    def set(
        self,
        params_value_new: xr.DataArray,
        params_name: str = "center",
        # params_attr: str = "value",
    ) -> xr.DataArray:
        return xr.apply_ufunc(
            self._set,
            self._obj,
            params_value_new,
            kwargs={
                "params_name": params_name,
                # "params_attr": params_attr,
            },
            input_core_dims=[[], ["params_dim"]],
            vectorize=True,
        )

    # TODO - Currently only sets values
    def _set(
        self,
        data: xr.DataArray,
        params_value_new: xr.DataArray,
        params_name: str = "center",
        # params_attr: str = "value",
    ):
        params = data.params
        pars = [key for key in params if key.endswith(params_name)]
        for i, par in enumerate(pars):
            data.params[par].set(value=params_value_new[i])
        return data



from typing import Literal

import numpy as np
import xarray as xr

from xrfit.base import DataArrayAccessor


@xr.register_dataarray_accessor("assess")
class AccessAccessor(DataArrayAccessor):
    def fit_stats(
        self,
        attr_name: Literal[
            "aic",
            "bic",
            "chisqr",
            "ci_out",
            "redchi",
            "rsquared",
            "success",
            "aborted",
            "ndata",
            "nfev",
            "nfree",
            "nvarys",
            "ier",
            "message",
        ] = "rsquared",
    ) -> xr.DataArray:
        def get_attr_safe(x, attr_name):
            try:
                return getattr(x, attr_name)
            except AttributeError:
                return np.nan

        return xr.apply_ufunc(
            get_attr_safe,
            self._obj,
            input_core_dims=[[]],
            kwargs={"attr_name": attr_name},
            vectorize=True,
            dask="parallelized",
        )

    def best_fit_stat(
        self,
        attr_name: Literal[
            "aic",
            "bic",
            "chisqr",
            "redchi",
            "rsquared",
        ] = "rsquared",
    ) -> dict:
        darr = self.fit_stats(attr_name)
        if attr_name in ["aic", "bic", "chisqr", "redchi"]:
            idx = np.unravel_index(darr.values.argmin(), darr.shape)
        if attr_name in ["rsquared"]:
            idx = np.unravel_index(darr.values.argmax(), darr.shape)
        return dict(zip(darr.dims, idx, strict=False))
        # return {
        #     dim: darr.coords[dim].values[i]
        #     for dim, i in zip(darr.dims, idx, strict=False)
        # }



import sys

import pyqtgraph as pg
import xarray as xr
from qtpy import QtCore, QtWidgets

from xrfit.base import DataArrayAccessor

# os.environ["QT_API"] = "pyqt6"
pg.setConfigOption("background", "w")
pg.setConfigOption("foreground", "k")


class MainWindow(QtWidgets.QWidget):
    def __init__(self, xarr) -> None:
        super().__init__()
        self._obj = xarr
        self.fit_stat = "rsquared"
        self.goodness_threshold_lower = 0.8
        self.goodness_threshold_upper = 1.5
        self.setWindowTitle("Display Manager")

        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)

        self.win = pg.GraphicsLayoutWidget()
        layout.addWidget(self.win)
        self.plot = self.win.addPlot(title="Fitting Result")
        initial_index = tuple([0] * (self._obj.ndim))
        x = self._obj[initial_index].item().userkws["x"]
        self.init_curve = self.plot.plot(
            x=x,
            y=self._obj[initial_index].item().init_fit,
            pen=pg.mkPen("b", width=3),
        )
        self.curve = self.plot.plot(
            x=x,
            y=self._obj[initial_index].item().best_fit,
            pen=pg.mkPen("r", width=3),
        )
        self.data_curve = self.plot.plot(
            x=x,
            y=self._obj[initial_index].item().data,
            symbol="o",
            pen=None,
            symbolBrush="k",
        )
        self.fix_ylim_checkbox = QtWidgets.QCheckBox("Fix Y-Axis Limits")
        self.fix_ylim_checkbox.toggled.connect(self.toggle_ylim)
        layout.addWidget(self.fix_ylim_checkbox)

        self.sliders = []
        self.slider_values = []
        self.slider_labels = []

        for dim in range(self._obj.ndim):
            slider_label = QtWidgets.QLabel(f"{self._obj.dims[dim]}: 0")
            slider = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal)
            slider.setMinimum(0)
            slider.setMaximum(self._obj.shape[dim] - 1)
            slider.valueChanged.connect(self.update_plot)
            self.sliders.append(slider)
            self.slider_values.append(0)
            self.slider_labels.append(slider_label)
            layout.addWidget(slider_label)
            layout.addWidget(slider)

        # Add dropdown for fit_stat
        self.fit_stat_dropdown = QtWidgets.QComboBox()
        self.fit_stat_dropdown.addItems(
            [
                "aic",
                "bic",
                "chisqr",
                "ci_out",
                "redchi",
                "rsquared",
                "success",
                "aborted",
                "ndata",
                "nfev",
                "nfree",
                "nvarys",
                "ier",
                "message",
            ]
        )
        self.fit_stat_dropdown.setCurrentText(self.fit_stat)
        self.fit_stat_dropdown.currentTextChanged.connect(self.update_fit_stat_label)
        layout.addWidget(QtWidgets.QLabel("Fit Statistic:"))
        layout.addWidget(self.fit_stat_dropdown)

        # Add a label to display the current fit_stat value
        self.fit_stat_label = QtWidgets.QLabel("Current Fit Stat: N/A")
        layout.addWidget(self.fit_stat_label)

        # Add a button to apply the input values
        self.apply_button = QtWidgets.QPushButton("Apply")
        self.apply_button.clicked.connect(self.apply_inputs)
        layout.addWidget(self.apply_button)
        # Add input fields for goodness_threshold_lower and goodness_threshold_upper
        self.goodness_threshold_lower_input = QtWidgets.QLineEdit(
            str(self.goodness_threshold_lower)
        )
        self.goodness_threshold_upper_input = QtWidgets.QLineEdit(
            str(self.goodness_threshold_upper)
        )
        layout.addWidget(QtWidgets.QLabel("Goodness Threshold Lower:"))
        layout.addWidget(self.goodness_threshold_lower_input)
        layout.addWidget(QtWidgets.QLabel("Goodness Threshold Upper:"))
        layout.addWidget(self.goodness_threshold_upper_input)

    def toggle_ylim(self, checked):
        if checked:
            y_range = self.plot.viewRange()[1]
            self.plot.enableAutoRange("y", False)
            self.plot.setYRange(y_range[0], y_range[1])
        else:
            self.plot.enableAutoRange("y", True)

    def update_plot(self, value):
        sender = self.sender()
        for i, slider in enumerate(self.sliders):
            if slider == sender:
                self.slider_values[i] = value
                self.slider_labels[i].setText(f"{self._obj.dims[i]}: {value}")

        index = tuple(self.slider_values)

        self.curve.setData(self._obj[index].item().best_fit)
        self.init_curve.setData(self._obj[index].item().init_fit)
        self.data_curve.setData(self._obj[index].item().data)

        self.update_slider_label_color(index)
        self.update_fit_stat_label(index)

    def update_slider_label_color(self, index):
        fit_stat = self._obj.assess.fit_stats(self.fit_stat)
        goodness_of_fit = fit_stat[index].item()
        for _, label in enumerate(self.slider_labels):
            if isinstance(goodness_of_fit, float):
                if (
                    self.goodness_threshold_lower
                    <= goodness_of_fit
                    <= self.goodness_threshold_upper
                ):
                    label.setStyleSheet("color: green;")
                else:
                    label.setStyleSheet("color: red;")

    def update_fit_stat_label(self, index=None):
        self.fit_stat = self.fit_stat_dropdown.currentText()
        if index is None:
            index = tuple(self.slider_values)
        fit_stat = self._obj.assess.fit_stats(self.fit_stat)
        try:
            current_fit_stat = fit_stat[index].item()
            self.fit_stat_label.setText(f"Current Fit Stat: {current_fit_stat}")
        except KeyError:
            self.fit_stat_label.setText("Current Fit Stat: N/A")

    def apply_inputs(self):
        self.fit_stat = self.fit_stat_dropdown.currentText()
        self.goodness_threshold_lower = float(
            self.goodness_threshold_lower_input.text()
        )
        self.goodness_threshold_upper = float(
            self.goodness_threshold_upper_input.text()
        )
        self.update_slider_label_color(tuple(self.slider_values))
        self.update_fit_stat_label(tuple(self.slider_values))


@xr.register_dataarray_accessor("display")
class DisplayAccessor(DataArrayAccessor):
    def __init__(self, xarray_obj):
        super().__init__(xarray_obj)

    def __call__(self):
        if not QtWidgets.QApplication.instance():
            qapp = QtWidgets.QApplication(sys.argv)
        else:
            qapp = QtWidgets.QApplication.instance()
        qapp.setStyle("Fusion")
        win = MainWindow(xarr=self._obj)
        win.show()
        win.activateWindow()
        qapp.exec()



from typing import Literal

import lmfit as lf
import numpy as np
import numpy.typing as npt
import xarray as xr

from xrfit.base import DataArrayAccessor


def _get_arr(
    model_result: lf.model.ModelResult,
    attr_name: Literal["best_fit", "init_fit", "residual", "data"] = "best_fit",
) -> npt.NDArray[np.float64]:
    # since parameter value might have changed, we re-evaluate the model just in case
    best_fit = model_result.eval()
    data = model_result.data
    if attr_name == "best_fit":
        return best_fit
    if attr_name == "init_fit":
        return model_result.init_fit
    if attr_name == "residual":
        return data.data - best_fit
    if attr_name == "data":
        return data
    raise ValueError(
        f"Invalid attr_name: {attr_name} (must be one of 'best_fit', 'init_fit', 'residual', 'data')"
    )


@xr.register_dataarray_accessor("get_arr")
class ArrAccessor(DataArrayAccessor):
    def _get_x(self):
        return self._obj[0].item().userkws["x"]

    def __call__(
        self,
        attr_name: Literal["best_fit", "init_fit", "residual", "data"] = "best_fit",
        new_dim_name: str = "x",
    ) -> xr.DataArray:
        return xr.apply_ufunc(
            _get_arr,
            self._obj,
            output_core_dims=[[new_dim_name]],
            kwargs={"attr_name": attr_name},
            vectorize=True,
            dask="parallelized",
        ).assign_coords(coords={new_dim_name: self._get_x()})
